/*
  XVarStar, GCVS catalogue explorer
  Copyright (C) 2001-2008 Ivan Vitjuk 

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <stdio.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <malloc.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/mman.h>

#include "variable.h"

#define RECORD_LEN              170   // bytes
#define STARS_NUMBER            31991  // number of stars in catalogue

/*
  types of fields in star record
 */
enum{
        TYPE_UNKNOWN,
        TYPE_INTEGER,
        TYPE_REAL,
        TYPE_STRING,
        TYPE_CHAR,
        TYPE_SIGN
};

/*
  holds informations about on field in star record
 */
struct GcvsData{
        unsigned int len, inc;
        void *data;
        int type;
};

/*
  works on first len chars in src

  - removes all head spaces
  - shrinks multiple spaces, tabs, etc. in src to one space in dst
  - removes tail spaces
 */
void shrink_spaces(char *dst, const char *src, size_t len)
{
        for(; isspace(*src) && len; src++, len--);

        while(*src && len--){
                if(!isspace(*src))
                        *dst++ = *src;
                else if(*(dst-1) != ' ')
                        *dst++ = ' ';
                src++;                
        }
        if(*(dst-1) == ' ')
                dst--;
        *dst = 0;
}

/*
  parse one line holding informations of star
 */
struct variable *_var_parse_line(const char *line)
{
        char buf[256];
        size_t offset, llen;
        struct variable *tmpvar;
        struct GcvsData *dptr;
        static struct variable var;
        static struct GcvsData gcvs_data[] = {
                {3, 3, &(var.constell), TYPE_INTEGER},
                {4, 4, &(var.number), TYPE_INTEGER},
                {1, 1, &(var.component), TYPE_CHAR},
                {10, 10, var.gcvs, TYPE_STRING},
                {1, 1, &(var.noteflag), TYPE_CHAR},
                {2, 2, &(var.Ra1950.h), TYPE_INTEGER},
                {2, 2, &(var.Ra1950.m), TYPE_INTEGER},
                {4, 4, &(var.Ra1950.s), TYPE_REAL},
                {3, 3, &(var.Dec1950.d), TYPE_INTEGER}, // two fields merged in one
                {2, 2, &(var.Dec1950.m), TYPE_INTEGER},
                {2, 2, &(var.Dec1950.s), TYPE_INTEGER},
                {1, 1, &(var.u_des), TYPE_CHAR},
                {10, 10, var.vartype, TYPE_STRING},
                {1, 1, &(var.l_magmax), TYPE_CHAR},
                {6, 6, &(var.magmax), TYPE_REAL},
                {1, 1, &(var.u_magmax), TYPE_CHAR},
                {1, 1, &(var.f_magmax), TYPE_CHAR},
                {1, 2, &(var.l_magmin), TYPE_CHAR},
                {6, 6, &(var.magmin), TYPE_REAL},
                {1, 1, &(var.u_magmin), TYPE_CHAR},
                {1, 1, &(var.n_magmin), TYPE_CHAR},
                {1, 1, &(var.f_magmin), TYPE_CHAR},
                {2, 2, var.magcode, TYPE_STRING},
                {14, 14, &(var.epoch), TYPE_REAL},
                {1, 1, &(var.q_epoch), TYPE_CHAR},
                {4, 4, &(var.yearnova), TYPE_INTEGER},
                {1, 1, &(var.l_period), TYPE_CHAR},
                {16, 16, &(var.period), TYPE_REAL},
                {1, 1, &(var.u_period), TYPE_CHAR},
                {3, 3, var.n_period, TYPE_STRING},
                {3, 3, var.mmd, TYPE_STRING},
                {1, 1, &(var.u_mmd), TYPE_CHAR},
                {1, 1, &(var.n_mmd), TYPE_CHAR},
                {17, 17, var.sptype, TYPE_STRING},
                {5, 6, var.ref1, TYPE_STRING},
                {5, 6, var.ref2, TYPE_STRING},
                {14, 14, var.exists, TYPE_STRING},
                {2, 2, &(var.Ra.h), TYPE_INTEGER},
                {2, 2, &(var.Ra.m), TYPE_INTEGER},
                {4, 4, &(var.Ra.s), TYPE_REAL},
                {3, 3, &(var.Dec.d), TYPE_INTEGER}, // two fields merged in one
                {2, 2, &(var.Dec.m), TYPE_INTEGER},
                {2, 2, &(var.Dec.s), TYPE_INTEGER},
                {0, 0, NULL, TYPE_UNKNOWN}          // end of data
        };

        if(!line)
                return NULL;

        if(!(tmpvar = (struct variable *)malloc(sizeof(*tmpvar))))
                return NULL;

        llen = strlen(line);

        memset(&var, 0, sizeof(var));

        /*
          walk through each field
          stop when end of line  or last field is reached
         */
        for(dptr=gcvs_data, offset=0; dptr->data; offset += dptr->inc, dptr++){
                if((offset + dptr->inc > llen+1) && dptr->type != TYPE_STRING)
                        break;

                switch(dptr->type){
                case TYPE_INTEGER:
                        memcpy(buf, line+offset, dptr->len);
                        buf[dptr->len] = '\0';
                        *((int *)(dptr->data)) = atoi(buf);
                        break;
                case TYPE_REAL:
                        memcpy(buf, line+offset, dptr->len);
                        buf[dptr->len] = '\0';
                        *((float *)(dptr->data)) = atof(buf);
                        break;
                case TYPE_CHAR:
                        *((char *)(dptr->data)) = *(line+offset);
                        if(!isprint(*((char *)(dptr->data))))
                                *((char *)(dptr->data)) = ' ';
                        break;
                case TYPE_STRING:
                        shrink_spaces(buf, line+offset, dptr->len);
                        memcpy((char *)(dptr->data), buf, strlen(buf));
                        break;
                default:
                        return NULL;
                }                                               
        }
        
        //printf("%d %d %d %s [%s]\n", 
        //     llen, var.constell, var.number, var.gcvs, line);
        
        *tmpvar = var;

        return tmpvar;
}

/*
  get next line from mmaped array of stars
 */
static int _var_load_next_line(char *dst, size_t dlen,
                               const char *src, size_t slen)
{
        static size_t offset=0;
        char *ptr;
        
        if(!dst || !src){
                offset = 0;
                return 1;
        }
        
        if(offset < slen){
                //memset(dst, 0, dlen);
                if((ptr = memccpy(dst, src+offset, '\n', dlen))){
                        *(ptr-1) = '\0';
                        offset += strlen(dst) + 1;
                }else{
                        printf("Database corruption\n");
                        exit(1);                               
                }
                //printf("%d: [%s]\n------------------------------\n", strlen(dst), dst);
                return 1;
        }
        return 0;
}

/*
  read stars line by line from GCVS catalogue
 */
static int _var_load_variables(variable_t *varlist, char *array, size_t alen)
{        
        struct variable *var;
        char buf[256];
        size_t counter=0;

        if(!array || !varlist){
                var_set_error(1);
                return 0;
        }
        
        while(_var_load_next_line(buf, sizeof(buf), array, alen)){
                counter++;
                if((counter%100) == 0){
                        var_draw_percent((float)counter/(float)STARS_NUMBER);
                }
                if((var = _var_parse_line(buf))){
                        var_list_insert(varlist, var);
                        free(var);
                }else{
                        var_list_free(varlist);
                        _var_load_next_line(NULL, 0, NULL, 0);
                        return 0;
                }
        }
        _var_load_next_line(NULL, 0, NULL, 0);

        return 1;
}

/*
  load stars from GCSV database in varlist  
 */
int var_load_variables(char *filename, variable_t *varlist)
{
        int fd, ret;
        struct stat statbuf;
        char *mmaparray;
        
        if(!filename || ! varlist){
                var_set_error(1);
                return 0;
        }
        
        if((fd = open(filename, O_RDONLY)) < 0){
                var_set_error(2);
                var_error_reason("open: %s: %s", filename, strerror(errno));
                return 0;
        }
        
        if(fstat(fd, &statbuf) < 0){
                var_set_error(2);
                var_error_reason("fstat: %s: %s", filename, strerror(errno));
                close(fd);
                return 0;
        }

        mmaparray = (char *)mmap((void *)NULL, (size_t)(statbuf.st_size), 
                                 PROT_READ, MAP_PRIVATE, fd, (off_t)0);
        
        if(!mmaparray){
                var_set_error(2);
                var_error_reason("mmap: %s: %s", filename, strerror(errno));
                close(fd);
                return 0;
        }
        
        ret = _var_load_variables(varlist, mmaparray, statbuf.st_size);
        
        munmap(mmaparray, statbuf.st_size);
        close(fd);

        return ret;
}
